#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "aclfs.h"

#ifndef SCM_CREDENTIALS
	#define SCM_CREDENTIALS 2
	// Struct not available in headers yet, but supported by the kernel
	struct ACL_UCRED {
		int32_t	pid;
		int32_t	uid;
		int32_t	gid;
	};
#else
	typedef struct ucred ACL_UCRED;
#endif


static int acl_fd=-1;

int acllib_connect ()
{
	int ret = -1;
	if (acl_fd != -1){
		ret = 0;
	}else{
		int fd =  socket (AF_UNIX,SOCK_STREAM,0);
		if (fd == -1){
			perror("socket client");
		}else{
			struct sockaddr_un un;
			un.sun_family = AF_UNIX;
			strcpy (un.sun_path,ACLFS_SOCKET);
			int s = connect(fd,(struct sockaddr*)&un,sizeof(un));
			if (s == -1){
				perror("connect");
			}else{
				acl_fd = fd;
				ret = 0;
			}
		}
	}
	return ret;
}


void acllib_sendreq (ACL_REQUEST &req, int lendata)
{
	char bufctl[sizeof(struct cmsghdr)+sizeof(ACL_UCRED)];
	struct cmsghdr *c = (struct cmsghdr*)bufctl;
	c->cmsg_len = CMSG_LEN(sizeof(ACL_UCRED));
	c->cmsg_level = SOL_SOCKET;
	c->cmsg_type = SCM_CREDENTIALS;
	ACL_UCRED *uc = (ACL_UCRED*)CMSG_DATA(c);
	uc->pid = getpid();
	uc->uid = getuid();
	uc->gid = getgid();


	struct msghdr h;

	struct iovec iov[1];
	iov[0].iov_base = &req;
	iov[0].iov_len = lendata + sizeof(ACL_REQUEST_BASE);

	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = (void*)bufctl;
	h.msg_controllen = sizeof(bufctl);
	h.msg_flags = 0;
	int ret = sendmsg (acl_fd,&h,0);
	if (ret==-1){
		perror ("sendmsg");
	}
}

/*
	Wait for a confirmation status.
	Return -1 if the original request failed.
*/
int acllib_receive()
{
	int ret = -1;
	union {
		char buf[100];
		ACL_REPLY rep;
	}u;
	int len = read (acl_fd,&u,sizeof(u));
	if (len == sizeof(ACL_REPLY)){
		if (u.rep.status == ACL_STATUS_OK){
			ret = 0;
		}else{
			errno = u.rep.errcode;
		}
	}
	return ret;
}

