#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <syslog.h>
#include "aclfsd.h"
#include <dlfcn.h>

static ACLMODULE **tbmod;
static int nbmod;
static int maxmod;

static void *dummy_initctx (const char *[])
{
	return NULL;
}

static void dummy_endctx (void *)
{
}


PUBLIC ACLMODULE::ACLMODULE(const char *_name)
{
	name = strdup(_name);
	char path[PATH_MAX];
	snprintf (path,sizeof(path)-1,"/usr/lib/aclfsd/modules/%s",_name);
	handle = dlopen (path,RTLD_LAZY);
	if (handle == NULL){
		syslog (LOG_ERR,"Can't load module %s: %m",_name);
	}else{
		initctx = (initctx_f)dlsym (handle,"initctx");
		if (initctx == NULL) initctx = dummy_initctx;
		endctx = (endctx_f)dlsym (handle,"endctx");
		if (endctx == NULL) endctx = dummy_endctx;
		testctx = (testctx_f)dlsym (handle,"testctx");
	}
}

PUBLIC ACLMODULE::~ACLMODULE()
{
	free (name);
	if (handle != NULL) dlclose (handle);
}

PUBLIC bool ACLMODULE::is_ok() const
{
	return handle != NULL;
}

ACLMODULE *aclmod_load (const char *name)
{
	ACLMODULE *ret = NULL;
	if (strchr(name,'/')==NULL){
		for (int i=0; i<nbmod; i++){
			if (strcmp(tbmod[i]->name,name)==0){
				ret = tbmod[i];
				break;
			}
		}
		if (ret == NULL){
			ret = new ACLMODULE (name);
			if (ret->is_ok()){
				if (nbmod == maxmod){
					maxmod += 100;
					tbmod = (ACLMODULE**)realloc(tbmod,maxmod*sizeof(ACLMODULE*));
					assert (tbmod != NULL);
				}
				tbmod[nbmod++] = ret;
			}else{
				delete ret;
				ret = NULL;
			}
		}
	}
	return ret;
}

