#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "remaudio.h"


static int clientsock_sinconnect (struct sockaddr_in &sin, int nbretry)
{
	int ret = -1;
	for (int i=0; i<nbretry; i++){		
		int s;
		if ((s = socket (AF_INET, SOCK_STREAM, 0)) >= 0) {
			if (connect (s, (struct sockaddr *) &sin, sizeof (sin))
				== -1){
				close (s);
				sleep(1);
			}else{
				ret = s;
				break;
			}
		}
	}
	return ret;
}

/*
	Obtient le port de communication via /etc/service
*/
static int clientsock_getport(const char *service)
{
	int ret = -1;
	struct servent *serv = getservbyname (service,"tcp");
	if (serv != NULL){
		ret = ntohs(serv->s_port);
	}
	return ret;
}


/*
	Etablie la connexion avec un serveur.
	Retourne -1 si erreur.
*/
int __clientsock_connect (const char *servname, int port, int nbretry)
{
	int ret = -1;
	struct sockaddr_in sin;
	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	if (__ipnum_validip (servname,true)){
		long addr = htonl(__ipnum_aip2l (servname));
		memcpy  (&sin.sin_addr,&addr, sizeof(addr));
		ret = clientsock_sinconnect (sin,nbretry);
	}else{
		struct hostent	*h = (struct hostent *) gethostbyname(servname);
		if (h == NULL){
			fprintf (stderr,"No server \"%s\" defined",servname);
		}else{
			memcpy  (&sin.sin_addr,h->h_addr, h->h_length);
			ret = clientsock_sinconnect (sin,nbretry);
		}
	}
	return ret;
}
/*
	Etablie la connexion avec un serveur.
	Retourne -1 si erreur.
*/
int __clientsock_connect (const char *servname, const char *portname, int nbretry)
{
	int port = clientsock_getport(portname);
	int ret = -1;
	if (port != -1){
		ret = __clientsock_connect (servname,port,nbretry);
	}
	return ret;
}

