#pragma implementation
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <wait.h>
#include <unistd.h>
#include <pwd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <stdarg.h>
#include "dispatch.h"

#define AUTOMOUNTOPT	0

#if AUTOMOUNTOPT
static AUTOMOUNT *autom = NULL;
#endif

static bool curpath_valid = false;
static char curpath[PATH_MAX];

const char *_dispatch_getcwd()
{
	if (!curpath_valid){
		curpath[0] = '\0';
		if (getcwd(curpath,sizeof(curpath))!=NULL){
			curpath_valid = true;
		}
	}
	return curpath;
}

const char *_dispatch_gethome()
{
	static const char *home = NULL;
//FILE *fout = fopen ("/tmp/d.log","a");
//if (fout != NULL){
//	fprintf (fout,"home=:%s: uid=%d %d\n",home,getuid(),geteuid());
//	fclose (fout);
//}
	if (home == NULL){
		if (getenv("VIRTUALFS_FROMUID")!=NULL){
			// Use eithr the uid or effective uid to tell which
			// user we are serving
			int uid = getuid();
			if (uid == 0){
				uid = geteuid();
			}
			if (uid != 0){
				struct passwd *p = getpwuid(uid);
				if (p != NULL){
					home = strdup(p->pw_dir);
				}else{
					home = strdup("/");
				}
			}
		}else{
			home = getenv("VIRTUALFS_HOME");
			if (home == NULL) home = getenv ("HOME");
			if (home == NULL || strcmp(home,"/")==0){
				// Compute the home using the uid
				int uid = getuid();
				struct passwd *p = getpwuid(uid);
				if (p != NULL){
					home = strdup(p->pw_dir);
				}else{
					home = strdup("/");
				}
			}
		}
		
	}
	return home;
}

/*
	Return the absolute path of the special directory n
*/
const char *_dispatch_getpathn(int &lenn)
{
	static char *pathn = NULL;
	static int len=0;
	if (pathn == NULL){
		const char *home = _dispatch_gethome();
		if (home != NULL){
			len = strlen(home)+3;
			pathn = (char*)malloc(len+1);
			sprintf (pathn,"%s/n/",home);
		}
	}
	lenn = len;
	return pathn;
}

int _dispatch_exec (const char *cmd)
{
	//fprintf (stderr,"exec: %s\n",cmd);
	if (__fork()==0){
		__chdir (_dispatch_gethome());
		char tmp[strlen(cmd)+1];
		strcpy (tmp,cmd);
		char *tb[20];
		int nb = 0;
		char *pt = tmp;
		while (*pt != '\0'){
			while (isspace (*pt)) pt++;
			if (*pt > ' '){
				char *start = pt;
				if (*pt == '"' || *pt == '\''){
					// Minimal quoting support
					char quote = *pt++;
					start = pt;
					while (*pt >= ' ' && *pt != quote) pt++;
				}else{
					while (*pt > ' ') pt++;
				}
				char last = *pt;
				*pt++ = '\0';
				tb[nb++] = start;
				if (last == '\0') break;
			}
		}
		tb[nb] = NULL;
		//for (int i=0; i<nb; i++) printf ("tb[%d]=:%s:\n",i,tb[i]);
		static const char *stdenv[]={
			"PATH=/bin:/usr/bin",
			NULL,
			NULL,
			NULL,
			NULL
		};
		char home[PATH_MAX];
		snprintf (home,sizeof(home)-1,"HOME=%s",_dispatch_gethome());
		char user[100];
		snprintf (user,sizeof(user)-1,"USER=%s",getenv("USER"));
		char display[300];
		snprintf (display,sizeof(display)-1,"DISPLAY=%s",getenv("DISPLAY"));
		stdenv[1] = home;
		stdenv[2] = user;
		stdenv[3] = display;
		char exepath[PATH_MAX];
		if (tb[0][0] != '/'){
			for (int i=0; i<3; i++){
				static const char *tbp[]={
					"/bin","/usr/bin","/usr/local/bin"
				};
				snprintf (exepath,sizeof(exepath),"%s/%s",tbp[i],tb[0]);
				struct stat st;
				if (stat(exepath,&st)!=-1){
					tb[0] = exepath;
					break;
				}
			}
		}
		__execve (tb[0],tb,(char**)stdenv);
	}
	int status;
	__wait (&status);
	//printf ("status = %d pid = %d\n",WEXITSTATUS(status),getpid());
	return WEXITSTATUS(status);
}

/*
	Simplify a path by stripping .. and .
	Return -1 if path is invalid, 0 if ok
	No validation are made on disk (for directory existence).

	If result is empty, replace by "." or "/" if the path was  absolute.
*/
int _dispatch_simplifypath (const char *path, char *newp)
{
	int ok = 0;
	char *debnewp = newp;
	char *lastvalid[20];	/* Last member other than . and .. */
	int nbvalid = 0;
	const bool isabs = path[0] == '/';
	char tmp[PATH_MAX];
	if (path == newp){
		strcpy (tmp,path);
		path = tmp;
	}
	while (path[0] == '.' && path[1] == '/') path += 2;
	if (path[0] != '/' && path[0] != '.'){
		lastvalid[nbvalid++] = newp;
	}
	while (ok == 0 && *path	!= '\0'){
		if (path[0] == '/'
			&& path[1] == '.'
			&& (path[2] == '/' || path[2] == '\0')){
			path += 2;
		}else if (path[0] == '/'
			&& path[1] == '.'
			&& path[2] == '.'
			&& (path[3] == '/' || path[3] == '\0')){
			if (nbvalid > 0){
				newp = lastvalid[--nbvalid];
				/* x/../y -> y and not /y */
				if (newp == debnewp
					&& !isabs
					&& path[3] == '/') path++;
			}else if (isabs){
				ok = -1;
				break;
			}else{
				*newp++ = '/';
				*newp++ = '.';
				*newp++ = '.';
				*newp = '\0';
			}
			path += 3;
		}else{
			if (*path == '/') lastvalid[nbvalid++] = newp;
			*newp++ = *path++;
		}
	}
	*newp = '\0';
	if (debnewp[0] == '\0'){
		debnewp[0] = isabs ? '/' : '.';
		debnewp[1] = '\0';
	}
	return ok;
}


#if AUTOMOUNTOPT

	static void checkmount (const char *path)
	{
		AUTOMOUNT *pt = autom;
		while (pt != NULL){
			if (pt->checkmount(path)!=-1) break;
			pt = pt->next;
		}
	}
#else
	#define checkmount(x)
#endif

static DISPATCH *first = NULL;
static DISPATCH *tbdispatch[1024];	// For now, 1024 file handles
static bool in_dispatch = false;

extern "C" int open (const char *fname, int mode, ...)
{
	va_list list;
	va_start (list,mode);
	int perm = va_arg(list,int);
	va_end (list);
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		#if 0
		{
			char tmp[1000];
			int len = sprintf (tmp,"dispatch open :%s: %d %d\n",fname,mode,perm);
			write (2,tmp,len);
		}
		#endif
		checkmount (fname);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->open != NULL){
				ret = pt->open (fname,mode,perm);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __libc_open (fname,mode,perm);
		}
		if (ret != -1){
			tbdispatch[ret] = pt;
		}
		in_dispatch = false;
	}else{
		ret = __libc_open (fname,mode,perm);
	}
	return ret;
}

extern "C" int open64 (const char *fname, int mode, ...)
{
	va_list list;
	va_start (list,mode);
	int perm = va_arg(list,int);
	va_end (list);
	// BUG 64
	return open (fname,mode,perm);
}


#if 0
extern "C" FILE *fopen (const char *fname, const char *fmode)
{
	FILE *ret = NULL;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (fname);

		const char *ptmode = fmode;
		int omode = 0, oflags = 0;
		char mode1 = *ptmode++;
		if (mode1 == 'r'){
			omode = O_RDONLY;
		}else if (mode1 == 'w'){
			omode = O_WRONLY;
			oflags = O_CREAT|O_TRUNC;
		}else if (mode1 == 'a'){
			omode = O_WRONLY;
			oflags = O_CREAT|O_APPEND;
		}
		if (ptmode[0] == '+' || (ptmode[0] == 'b' && ptmode[1] == '+')) omode = O_RDWR;
		int mode = omode | oflags;
		const int perm = 0666;

		#if 0
		{
			char tmp[1000];
			int len = sprintf (tmp,"dispatch fopen :%s: :%s: -> %d %d\n",fname,fmode,mode,perm);
			write (2,tmp,len);
		}
		#endif



		DISPATCH *pt = first;
		int handle = DISPATCH_DONTCARE;
		while (pt != NULL){
			if (pt->open != NULL){
				handle = pt->open (fname,mode,perm);
				if (handle != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		#if 0
		{
			char tmp[1000];
			int len = sprintf (tmp,"dispatch fopen handle = %d\n",handle);
			write (2,tmp,len);
		}
		#endif
		if (handle == DISPATCH_DONTCARE){
			ret = _IO_fopen (fname,fmode);
			if (ret != NULL){
				tbdispatch[fileno(ret)] = NULL;
			}
		}else if (handle != -1){
			ret = fdopen (handle,fmode);
			tbdispatch[handle] = pt;
			if (ret == NULL){
				close (handle);
			}
		}
		#if 0
		{
			char tmp[1000];
			int len = sprintf (tmp,"dispatch fopen ret = %p\n",ret);
			write (2,tmp,len);
		}
		#endif
		in_dispatch = false;
	}else{
		ret = _IO_fopen (fname,fmode);
	}
	return ret;
}

extern "C" FILE *freopen (const char *fname, const char *fmode, FILE *f)
{
	fprintf (stderr,"freopen %s\n",fname);
	#if 0
		// Difficult to achieve this one cleanly
		FILE *newf = fopen (fname,mode);
		if (newf != NULL){
			
		}else{
			fclose (f);
			f = NULL;
		}
	#endif
	return NULL;
}


extern "C" FILE *fopen64 (const char *fname, const char *fmode)
{
	return fopen (fname,fmode);
}

extern "C" FILE *freopen64 (const char *fname, const char *fmode, FILE *f)
{
	return freopen (fname,fmode,f);
}


extern "C" size_t fread (void *p, size_t s, size_t nb, FILE *fin)
{
	int ret = -1;
	int handle = fileno(fin);
	DISPATCH *pt = tbdispatch[handle];
	if (pt != NULL && pt->read != NULL){
		ret = pt->read (handle,p,s*nb);
		if (ret > 0) ret /= s;
	}else{
		ret = _IO_fread (p,s,nb,fin);
	}
	return ret;
}

extern "C" size_t fwrite (const void *p, size_t s, size_t nb, FILE *fin)
{
	int ret = -1;
	int handle = fileno(fin);
	DISPATCH *pt = tbdispatch[handle];
	if (pt != NULL && pt->write != NULL){
		ret = pt->write (handle,p,s*nb);
		if (ret > 0) ret /= s;
	}else{
		ret = _IO_fwrite (p,s,nb,fin);
	}
	return ret;
}

extern "C" int fseek (FILE *f, long off, int whence)
{
	int ret = -1;
	int handle = fileno(f);
	DISPATCH *pt = tbdispatch[handle];
	if (pt != NULL && pt->lseek != NULL){
		ret = pt->lseek (handle,off,whence);
	}else{
		#if 0
			#ifndef _IO_pos_BAD
				#define _IO_pos_BAD ((_IO_fpos64_t)(-1))
			#endif
			#define _IO_fseek(__fp, __offset, __whence) \
			  (_IO_seekoff(__fp, __offset, __whence, _IOS_INPUT|_IOS_OUTPUT) == _IO_pos_BAD ? EOF : 0)
		#endif
		ret = __fseek (f,off,whence);
	}
	return ret;
}

extern "C" void rewind (FILE *f)
{
	int handle = fileno(f);
	DISPATCH *pt = tbdispatch[handle];
	if (pt != NULL && pt->lseek != NULL){
		// Possible bug.
		pt->lseek (handle,0,SEEK_SET);
	}else{
		__rewind (f);
	}
}
#endif

#if 0
// creat is calling open underneath, so no need to override it
// The symbol is not weak anyway.
extern "C" int creat (const char *fname, mode_t mode)
{
	checkmount (fname);
	if (state == OPEN_UNKNOWN){
		netaudio_setstate();
	}
	int ret = -1;
	if (state == OPEN_REMOTE){
		if (strcmp(fname,"/dev/fd0")==0
			|| strcmp(fname,"/dev/fd0H1440")==0
			|| strcmp(fname,"/dev/fd1")==0
			|| strcmp(fname,"/dev/fd1H1440")==0){
			ret = netaudio_connect (fname,mode,NET_FLOPPY);
		}
	}
	if (ret == -1){
		ret = __libc_creat (fname,mode);
	}
	return ret;
}
#endif

extern "C" off_t lseek(int fd, off_t offset, int whence)
{
	DISPATCH *pt = tbdispatch[fd];
	off_t ret = 0;
	if (pt != NULL && pt->lseek != NULL){
		ret = pt->lseek (fd,offset,whence);
	}else{
		ret = __libc_lseek (fd,offset,whence);
	}
	return ret;
}
#if __GLIBC__ > 2 || ( __GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)
extern "C" __off64_t lseek64(int fd, __off64_t offset, int whence)
{
	DISPATCH *pt = tbdispatch[fd];
	__off64_t ret = 0;
	// fprintf (stderr,"lseek64 %ld\n",offset);
	if (pt != NULL && pt->lseek != NULL){
		ret = pt->lseek (fd,offset,whence);
	}else{
		ret = __lseek64 (fd,offset,whence);
	}
	return ret;
}
#endif

extern "C" int close (int fd)
{
	int ret = -1;
	if (fd != -1){
		DISPATCH *pt = tbdispatch[fd];
		ret = 0;
		if (pt != NULL && pt->close != NULL){
			ret = pt->close (fd);
		}else{
			ret = __libc_close (fd);
		}
		tbdispatch[fd] = NULL;
	}
	return ret;
}

extern "C" int write (int fd, const void *buf, unsigned len)
{
	DISPATCH *pt = tbdispatch[fd];
	int ret = 0;
	if (pt != NULL && pt->write != NULL){
		ret = pt->write (fd,buf,len);
	}else{
		ret = __libc_write (fd,buf,len);
	}
	return ret;
}

extern "C" int read (int fd, void *buf, unsigned len)
{
	DISPATCH *pt = tbdispatch[fd];
	int ret = 0;
	if (pt != NULL && pt->read != NULL){
		ret = pt->read (fd,buf,len);
	}else{
		ret = __libc_read (fd,buf,len);
	}
	return ret;
}

extern "C" int ioctl (int fd, unsigned request, void *a1, void *a2, void *a3)
{
	DISPATCH *pt = tbdispatch[fd];
	int ret = 0;
	if (pt != NULL && pt->ioctl != NULL){
		ret = pt->ioctl (fd,request,a1,a2,a3);
	}else{
		ret = __ioctl (fd,request,a1,a2,a3);
	}
	return ret;
}

extern "C"{
	int __vfs_xstat (int ver, const char *fname, struct stat *buf);
	int __vfs_lxstat (int ver, const char *fname, struct stat *buf);
	int __vfs_xstat64 (int ver, const char *fname, struct stat64 *buf);
	int __vfs_lxstat64 (int ver, const char *fname, struct stat64 *buf);
}

extern "C" int _vfs_xstat (int ver, const char *fname, struct stat *buf)
{
	// vfs_xstat is a special trick. Many package are compiled
	// so they used the inline version of the stat and lstat function.
	// Those inline function are calling __xstat and __lxstat directly
	// instead of the weak symbol, so they can't be overriden.
	// The glibc-vfs package provides a fix so newly compiled package
	// will call the proper weak symbol.
	//
	// In the mean time (I do not want people to recompile everything
	// just to test), I have modified glibc so __xstat and __xlstat
	// are calling a weak symbol _vfs_xstat and _vfs_lxstat. This
	// weak symbol point to __vfs_xstat and __vfs_lxstat.
	//
	// With this, I can override _vfs_xstat. One may note that
	// the overriden (below) _xstat function is still calling __xstat
	// which then will call this overriden function _vfs_xstat which
	// ultimatly will call __vfs_xstat. Why not calling __vfs_xstat
	// directly.
	//
	// At some point, I expect to see the change in glibc go in
	// the standard glibc (and this even if virtualfs is a failure)
	// But the vfs_xstat hack won't be need anymore. I want people using
	// virtualfs to be able to upgrade to newer glibc. If they upgrade
	// to a fixed glibc, without the vfs hack, virtualfs will still operate
	// and this overriden _vfs_xstat function won't be called anymore.
	int ret = DISPATCH_DONTCARE;
	// fprintf (stderr,"_vfs_xstat :%s:\n",fname);
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (fname);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->stat != NULL){
				ret = pt->stat (ver,fname,buf);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __vfs_xstat (ver,fname,buf);
		}
		in_dispatch = false;
	}else{
		ret = __vfs_xstat (ver,fname,buf);
	}
	return ret;
}

#if __GLIBC__ > 2 || ( __GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)
struct stat64 {
    __dev_t st_dev;			/* Device.  */
    unsigned short int __pad1;

    __ino64_t st_ino;			/* File serial number.	*/
    __mode_t st_mode;			/* File mode.  */
    __nlink_t st_nlink;			/* Link count.  */
    __uid_t st_uid;			/* User ID of the file's owner.	*/
    __gid_t st_gid;			/* Group ID of the file's group.*/
    __dev_t st_rdev;			/* Device number, if device.  */
    unsigned short int __pad2;
    __off64_t st_size;			/* Size of file, in bytes.  */
    unsigned long int st_blksize;	/* Optimal block size for I/O.  */

    __blkcnt64_t st_blocks;		/* Number 512-byte blocks allocated. */
    __time_t st_atime;			/* Time of last access.  */
    unsigned long int __unused1;
    __time_t st_mtime;			/* Time of last modification.  */
    unsigned long int __unused2;
    __time_t st_ctime;			/* Time of last status change.  */
    unsigned long int __unused3;
    unsigned long int __unused4;
    unsigned long int __unused5;
};
struct dirent_64{
	__ino64_t d_ino;
	__off64_t d_off;
	unsigned short int d_reclen;
	unsigned char d_type;
	char d_name[256];		/* We must not include limits.h! */
};

static void dispatch_stat_to_stat64(
	struct stat &buf32,
	struct stat64 &buf)
{
	// fprintf (stderr,"convert stat %d %d\n",sizeof(struct stat),sizeof(struct stat64));
	buf.st_dev = buf32.st_dev;
	buf.st_ino = buf32.st_ino;
	buf.st_mode = buf32.st_mode;
	buf.st_nlink = buf32.st_nlink;
	buf.st_uid = buf32.st_uid;
	buf.st_gid = buf32.st_gid;
	buf.st_rdev = buf32.st_rdev;
	buf.st_size = buf32.st_size;
	buf.st_blksize = buf32.st_blksize;
	buf.st_blocks = buf32.st_blocks;
	buf.st_atime = buf32.st_atime;
	buf.st_mtime = buf32.st_mtime;
	buf.st_ctime = buf32.st_ctime;
}



extern "C" int _vfs_xstat64 (int ver, const char *fname, struct stat64 *buf)
{
	int ret = DISPATCH_DONTCARE;
	// fprintf (stderr,"_vfs_xstat64 :%s:\n",fname);
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (fname);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->stat != NULL){
				struct stat buf32;
				ret = pt->stat (ver,fname,&buf32);
				if (ret != DISPATCH_DONTCARE){
					/* #Specification: 64 bits services / stat
						Currently, the dispatch interface only
						support standard services (32 bits) and
						we map 64 bits services to 32. Temporary
						mesure to simplify operation on both
						glibc 2.0 and 2.1.
					*/
					dispatch_stat_to_stat64 (buf32,*buf);
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __vfs_xstat64 (ver,fname,buf);
		}
		in_dispatch = false;
	}else{
		ret = __vfs_xstat64 (ver,fname,buf);
	}
	return ret;
}

extern "C" int _vfs_lxstat64 (int ver, const char *fname, struct stat64 *buf)
{
	int ret = DISPATCH_DONTCARE;
	// fprintf (stderr,"_vfs_lxstat64 :%s:\n",fname);
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (fname);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->stat != NULL){
				struct stat buf32;
				ret = pt->lstat (ver,fname,&buf32);
				if (ret != DISPATCH_DONTCARE){
					dispatch_stat_to_stat64 (buf32,*buf);
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __vfs_lxstat64 (ver,fname,buf);
		}
		in_dispatch = false;
	}else{
		ret = __vfs_lxstat64 (ver,fname,buf);
	}
	return ret;
}
#endif

extern "C" int _xstat (int ver, const char *fname, struct stat *buf)
{
	// fprintf (stderr,"_xstat :%s:\n",fname);
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (fname);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->stat != NULL){
				ret = pt->stat (ver,fname,buf);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __xstat (ver,fname,buf);
		}
		in_dispatch = false;
	}else{
		ret = __xstat (ver,fname,buf);
	}
	return ret;
}

extern "C" int _vfs_lxstat (int ver, const char *fname, struct stat *buf)
{
	// fprintf (stderr,"_vfs_lxstat :%s: %d\n",fname,in_dispatch);
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (fname);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->lstat != NULL){
				ret = pt->lstat (ver,fname,buf);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __vfs_lxstat (ver,fname,buf);
		}
		in_dispatch = false;
	}else{
		ret = __vfs_lxstat (ver,fname,buf);
	}
	return ret;
}


extern "C" int _lxstat (int ver, const char *fname, struct stat *buf)
{
	// fprintf (stderr,"_lxstat :%s:\n",fname);
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (fname);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->lstat != NULL){
				ret = pt->lstat (ver,fname,buf);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __lxstat (ver,fname,buf);
		}
		in_dispatch = false;
	}else{
		ret = __lxstat (ver,fname,buf);
	}
	return ret;
}
extern "C" int _fxstat (int ver, int fd, struct stat *buf)
{
	DISPATCH *pt = tbdispatch[fd];
	int ret = 0;
	if (pt != NULL && pt->fstat != NULL){
		ret = pt->fstat (ver,fd,buf);
		//fprintf (stderr,"fstat %d\n",fd);
	}else{
		ret = __fxstat (ver,fd,buf);
	}
	return ret;
}

extern "C" int readlink (const char *path, char *buf, size_t bufsiz)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->readlink != NULL){
				ret = pt->readlink (path,buf,bufsiz);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __readlink (path,buf,bufsiz);
		}
		in_dispatch = false;
	}else{
		ret = __readlink (path,buf,bufsiz);
	}
	return ret;
}

#if 0
extern "C" int utimes (const char *path, struct timeval *tvp)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->utimes != NULL){
				ret = pt->utimes (path,tvp);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __utimes (path,tvp);
		}
		in_dispatch = false;
	}else{
		ret = __utimes (path,tvp);
	}
	return ret;
}
#endif

extern "C" int utime (const char *path, const struct utimbuf *tvp)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->utime != NULL){
				ret = pt->utime (path,tvp);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __utime (path,tvp);
		}
		in_dispatch = false;
	}else{
		ret = __utime (path,tvp);
	}
	return ret;
}

extern "C" int access (const char *path, int type)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->access != NULL){
				ret = pt->access (path,type);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __access (path,type);
		}
		in_dispatch = false;
	}else{
		ret = __access (path,type);
	}
	return ret;
}

extern "C" int euidaccess (const char *path, int type)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->euidaccess != NULL){
				ret = pt->euidaccess (path,type);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __euidaccess (path,type);
		}
		in_dispatch = false;
	}else{
		ret = __euidaccess (path,type);
	}
	return ret;
}

extern "C" int chdir (const char *path)
{
	//fprintf (stderr,"chdir %s\n",path);
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->chdir != NULL){
				ret = pt->chdir (path);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __chdir (path);
		}
		in_dispatch = false;
	}else{
		ret = __chdir (path);
	}
	curpath_valid = false;
	#if AUTOMOUNTOPT
		AUTOMOUNT *pt = autom;
		while (pt != NULL){
			if (pt->afterchdir != NULL) pt->afterchdir(path);
			pt = pt->next;
		}
	#endif
	return ret;
}


extern "C" int chmod (const char *path, mode_t mode)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->chmod != NULL){
				ret = pt->chmod (path,mode);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __chmod (path,mode);
		}
		in_dispatch = false;
	}else{
		ret = __chmod (path,mode);
	}
	return ret;
}
extern "C" int fchmod (int fd, mode_t mode)
{
	DISPATCH *pt = tbdispatch[fd];
	int ret = 0;
	if (pt != NULL && pt->fchmod != NULL){
		ret = pt->fchmod (fd,mode);
	}else{
		ret = __fchmod (fd,mode);
	}
	return ret;
}
extern "C" int fchown (int fd, uid_t uid, gid_t gid)
{
	DISPATCH *pt = tbdispatch[fd];
	int ret = 0;
	if (pt != NULL && pt->fchown != NULL){
		ret = pt->fchown (fd,uid,gid);
	}else{
		ret = __fchown (fd,uid,gid);
	}
	return ret;
}

extern "C" int chown (const char *path, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->chown != NULL){
				ret = pt->chown (path,uid,gid);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __chown (path,uid,gid);
		}
		in_dispatch = false;
	}else{
		ret = __chown (path,uid,gid);
	}
	return ret;
}

extern "C" int lchown (const char *path, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->lchown != NULL){
				ret = pt->lchown (path,uid,gid);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __lchown (path,uid,gid);
		}
		in_dispatch = false;
	}else{
		ret = __lchown (path,uid,gid);
	}
	return ret;
}

struct DIR_LOOKUP {
	DIR *dir;
	DISPATCH *pt;
};

static DIR_LOOKUP tbdir[100];
static int maxdir = 0;
	

extern "C" DIR *opendir (const char *path)
{
	//fprintf (stderr,"opendir %s\n",path);
	DIR *ret = NULL;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->opendir != NULL){
				ret = pt->opendir (path);
				if (ret != NULL){
					// We register this DIR object with the DISPATH handler
					// so readdir and closedir can tell if they are working
					// with a vfs or not
					DIR_LOOKUP *ptdir = tbdir;
					for (int i=0; i<100; i++, ptdir++){
						if (ptdir->pt == NULL){
							ptdir->pt = pt;
							ptdir->dir = ret;
							if (i >= maxdir) maxdir = i+1;
							break;
						}
					}
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == NULL){
			ret = __opendir (path);
		}
		in_dispatch = false;
	}else{
		ret = __opendir (path);
	}
	return ret;
}

extern "C" int closedir (DIR *dir)
{
	//fprintf (stderr,"closedir\n");
	int ret = -1;
	DIR_LOOKUP *ptdir = tbdir;
	int i;
	for (i=0; i<maxdir; i++, ptdir++){
		if (ptdir->dir == dir){
			ret = ptdir->pt->closedir (dir);
			ptdir->pt = NULL;
			ptdir->dir = NULL;
			break;
		}
	}
	if (i == maxdir) ret = __closedir (dir);
	return ret;
}

extern "C" struct dirent *readdir (DIR *dir)
{
	struct dirent *ret = NULL;
	DIR_LOOKUP *ptdir = tbdir;
	int i;
	for (i=0; i<maxdir; i++, ptdir++){
		if (ptdir->dir == dir){
			ret = ptdir->pt->readdir (dir);
			break;
		}
	}
	if (i == maxdir) ret = __readdir (dir);
	//if (ret != NULL) fprintf (stderr,"readdir %s\n",ret->d_name);
	return ret;
}

extern "C" ssize_t getdirentries(
	int fd,
	char *buf,
	size_t  nbytes,
	off_t *basep)
{
	fprintf (stderr,"getdirentries %d\n",fd);
	return __getdirentries (fd,buf,nbytes,basep);
}


#if __GLIBC__ > 2 || ( __GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)

/*
	For now readdir64 is simply mapped to normal readdir
*/
extern "C" struct dirent_64 *readdir64 (DIR *dir)
{
	struct dirent *ret = NULL;
	DIR_LOOKUP *ptdir = tbdir;
	int i;
	for (i=0; i<maxdir; i++, ptdir++){
		if (ptdir->dir == dir){
			ret = ptdir->pt->readdir (dir);
			break;
		}
	}
	if (i == maxdir) ret = __readdir (dir);
	struct dirent_64 *ret64 = NULL;
	if (ret != NULL){
		//fprintf (stderr,"readdir64 %s\n",ret->d_name);
		static struct dirent_64 s;
		s.d_ino = ret->d_ino;
		s.d_off = ret->d_off;
		s.d_type = ret->d_type;
		s.d_reclen = ret->d_reclen;
		memcpy (s.d_name,ret->d_name,s.d_reclen+1);
		ret64 = &s;
	}
	return ret64;
}
#endif

extern "C" int rename (const char *old, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (old);
		checkmount (newp);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->rename != NULL){
				ret = pt->rename (old,newp);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __rename (old,newp);
		}
		in_dispatch = false;
	}else{
		ret = __rename (old,newp);
	}
	return ret;
}

extern "C" int unlink (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->unlink != NULL){
				ret = pt->unlink (path);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __unlink (path);
		}
		in_dispatch = false;
	}else{
		ret = __unlink (path);
	}
	return ret;
}

extern "C" int link (const char *oldpath, const char *newpath)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (oldpath);
		checkmount (newpath);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->link != NULL){
				ret = pt->link (oldpath,newpath);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __link (oldpath,newpath);
		}
		in_dispatch = false;
	}else{
		ret = __link (oldpath,newpath);
	}
	return ret;
}

extern "C" int symlink (const char *oldpath, const char *newpath)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (newpath);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->symlink != NULL){
				ret = pt->symlink (oldpath,newpath);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __symlink (oldpath,newpath);
		}
		in_dispatch = false;
	}else{
		ret = __symlink (oldpath,newpath);
	}
	return ret;
}

extern "C" int mkdir (const char *path, mode_t mode)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->mkdir != NULL){
				ret = pt->mkdir (path,mode);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __mkdir (path,mode);
		}
		in_dispatch = false;
	}else{
		ret = __mkdir (path,mode);
	}
	return ret;
}

extern "C" int rmdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (path);
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->rmdir != NULL){
				ret = pt->rmdir (path);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __rmdir (path);
		}
		in_dispatch = false;
	}else{
		ret = __rmdir (path);
	}
	return ret;
}

extern "C" int execve (
	const char *filename,
	char *const argv[],
	char *const envp[])
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		checkmount (filename);
		#if 0
			fprintf (stderr,"execve :%s:",filename);
			for (int i=0; argv[i] != NULL; i++) fprintf (stderr," [%s]",argv[i]);
			fprintf (stderr,"\n");
			for (int i=0; envp[i] != NULL; i++) fprintf (stderr," <%s>",envp[i]);
			fprintf (stderr,"\n");
		#endif
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->execve != NULL){
				ret = pt->execve (filename,argv,envp);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		in_dispatch = false;
		if (ret == DISPATCH_DONTCARE){
			ret = __execve (filename,argv,envp);
		}
	}else{
		ret = __execve (filename,argv,envp);
	}
	return ret;
}



extern "C" int socket (int domain, int type, int protocol)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->socket != NULL){
				ret = pt->socket (domain,type,protocol);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __socket (domain,type,protocol);
		}
		in_dispatch = false;
	}else{
		ret = __socket (domain,type,protocol);
	}
	return ret;
}

extern "C" int connect (int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->connect != NULL){
				ret = pt->connect (sockfd,serv_addr,addrlen);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __connect (sockfd,serv_addr,addrlen);
		}
		in_dispatch = false;
	}else{
		ret = __connect (sockfd,serv_addr,addrlen);
	}
	return ret;
}

#if 0
extern "C" int bind (int sockfd, const struct sockaddr *my_addr, socklen_t addrlen)
{
	int ret = DISPATCH_DONTCARE;
	if (!in_dispatch){
		in_dispatch = true;
		DISPATCH *pt = first;
		while (pt != NULL){
			if (pt->bind != NULL){
				ret = pt->bind (sockfd,my_addr,addrlen);
				if (ret != DISPATCH_DONTCARE){
					break;
				}
			}
			pt = pt->next;
		}
		if (ret == DISPATCH_DONTCARE){
			ret = __bind (sockfd,my_addr,addrlen);
		}
		in_dispatch = false;
	}else{
		ret = __bind (sockfd,my_addr,addrlen);
	}
	return ret;
}
#endif


REGISTER_DISPATCH::REGISTER_DISPATCH( DISPATCH &m)
{
	// Put at the end of the list to respect the LD_PRELOAD order
	DISPATCH **ptpt = &first;
	while (*ptpt != NULL){
		ptpt = &(*ptpt)->next;
	}
	*ptpt = &m;
	m.next = NULL;
}

#if AUTOMOUNTOPT

REGISTER_AUTOMOUNT::REGISTER_AUTOMOUNT( AUTOMOUNT &m)
{
	// Put at the end of the list to respect the LD_PRELOAD order
	AUTOMOUNT **ptpt = &autom;
	while (*ptpt != NULL){
		ptpt = &(*ptpt)->next;
	}
	*ptpt = &m;
	m.next = NULL;
}
	
#endif
