#pragma interface
#ifndef DISPATCH_H
#define DISPATCH_H

#include <unistd.h>
#include <sys/types.h>
#include <utime.h>
#include <sys/socket.h>

#ifdef __GLIBC__
	#if __GLIBC__ > 2 || ( __GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)
		extern "C" __off64_t __lseek64 (int __fd, __off64_t __offset, int __whence);
	#endif
#else
	typedef int socklen_t;
#endif

extern "C" {
	int __libc_open (const char *, int, int);
	int __open64 (const char *, int, int);
	int __libc_creat (const char *, int);
	int __libc_close(int);
	int __libc_write (int, const void *, unsigned);
	int __libc_read  (int, void *, unsigned);
	int __ioctl (int, int request, ...);
	off_t __libc_lseek(int fd, off_t offset, int whence);
	int __xstat (int ver, const char *fname, struct stat *buf);
	int __fxstat (int ver, int fd, struct stat *buf);
	int __lxstat (int ver, const char *fname, struct stat *buf);
	int __chdir (const char *path);
	struct DIR *__opendir (const char *path);
	struct dirent *__readdir (struct DIR *dir);
	int __closedir (struct DIR *dir);
	int __rename (const char *old, const char *newp);
	int __unlink (const char *path);
	int __rmdir (const char *path);
	int __mkdir (const char *path, mode_t mode);
	int __socket (int domain, int type, int protocol);
	int __bind (int sockfd, const struct sockaddr *my_addr, socklen_t addrlen);
	int __connect (int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen);
	int __chmod (const char *path, mode_t mode);
	int __chown (const char *path, uid_t uid, gid_t gid);
	int __lchown (const char *path, uid_t uid, gid_t gid);
	FILE *_IO_fopen (const char *fname, const char *mode);
	int __readlink (const char *path, char *buf, size_t bufsiz);
	int __utimes (const char *filename, struct timeval *tvp);
	int __utime (const char *filename, const struct utimbuf *tvp);
	int __access (const char *file, int type);
	int __euidaccess (const char *file, int type);
	int _IO_fread (void *p, size_t s, size_t n, FILE *f);
	int _IO_fwrite (const void *p, size_t s, size_t n, FILE *f);
	int __fseek (FILE *f, long offset, int whence);
	long _IO_ftell (FILE *f);
	void __rewind (FILE *f);
	int __execve (const char *, char *const argv[], char *const envp[]);
	int __link (const char *, const char *);
	int __symlink (const char *, const char *);
	ssize_t __getdirentries(int fd, char *buf, size_t  nbytes  , off_t *basep);
	int __fchmod(int, mode_t);
	int __fchown(int, uid_t, gid_t);
}


struct DISPATCH {
	DISPATCH *next;
	int (*open) (const char *, int, int);
	int (*close) (int);
	int (*write) (int, const void *, unsigned);
	int (*read)  (int, void *, unsigned);
	int (*ioctl) (int, unsigned request, void *a1, void *a2, void *a3);
	off_t (*lseek) (int fd, off_t offset, int whence);
	int (*stat) (int ver, const char *fname, struct stat *buf);
	int (*fstat) (int ver, int fd, struct stat *buf);
	int (*lstat) (int ver, const char *fname, struct stat *buf);
	int (*chdir) (const char *path);
	struct DIR * (*opendir) (const char *path);
	struct dirent *(*readdir) (struct DIR *dir);
	int (*closedir) (struct DIR *dir);
	int (*rename) (const char *old, const char *newp);
	int (*unlink) (const char *path);
	int (*rmdir) (const char *path);
	int (*mkdir) (const char *path, int mode);
	int (*socket) (int domain, int type, int protocol);
	int (*connect) (int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen);
	int (*bind) (int sockfd, const struct sockaddr *my_addr, socklen_t addrlen);
	int (*chmod) (const char *path, mode_t mode);
	int (*chown) (const char *path, uid_t uid, gid_t gid);
	int (*lchown) (const char *path, uid_t uid, gid_t gid);
	int (*readlink) (const char *path, char *buf, size_t bufsiz);
	int (*utime) (const char *filename, const struct utimbuf *tvp);
	int (*access) (const char *file, int type);
	int (*euidaccess) (const char *file, int type);
	int (*link) (const char *, const char *);
	int (*symlink) (const char *, const char *);
	int (*fchmod)(int, mode_t);
	int (*fchown) (int, uid_t, gid_t);
	int (*execve) (const char *, char * const [], char *const[]);
};

struct AUTOMOUNT{
	AUTOMOUNT *next;
	int (*checkmount) (const char *path);
	void (*afterchdir) (const char *path);
};

class REGISTER_DISPATCH{
public:
	REGISTER_DISPATCH(DISPATCH &);
};

class REGISTER_AUTOMOUNT{
public:
	REGISTER_AUTOMOUNT(AUTOMOUNT &);
};

#define DISPATCH_DONTCARE	-10000

const char *_dispatch_getcwd();
const char *_dispatch_gethome();
const char *_dispatch_getpathn(int &lenn);
int _dispatch_exec (const char *);
int _dispatch_simplifypath (const char *path, char *newp);

#endif

