/*
	Few basic file system tests
	Execyte without argument to see the usage screen
*/
//#define __USE_FILE_OFFSET64
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
//#define __USE_LARGEFILE64
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/sysmacros.h>
#include <utime.h>
#include <errno.h>

static void t_error (const char *ctl, ...)
{
	va_list list;
	va_start (list,ctl);
	vfprintf (stderr,ctl,list);
	fprintf (stderr," %d(%s)\n",errno,strerror(errno));
	exit (-1);
}

static void t_success()
{
	printf ("**** Success\n");
}

static void t_exist (const char *fname, bool exist)
{
	struct stat st;
	int ret = stat (fname,&st);
	if (exist){
		if (ret != 0) t_error ("t_exist: Missing file %s",fname);
	}else{
		if (ret == 0) t_error ("t_exist: Unexpected file %s",fname);
	}
}

static void t_size (const char *fname, int expected)
{
	struct stat st;
	int ret = stat (fname,&st);
	if (ret != 0){
		t_error ("t_size: missing file %s",fname);
	}else if (st.st_size != expected){
		t_error ("t_size: file %s wrong size %d, expected %d",fname
			,st.st_size,expected);
	}
}
static void t_fopen (const char *fname)
{
	printf ("Test fopen: %s\n",fname);
	FILE *fout = fopen (fname,"w");
	if (fout == NULL){
		t_error ("fopen(%s w)",fname);
	}else{
		t_exist (fname,true);
		fwrite ("hello\n",1,6,fout);
		fflush (fout);
		t_size (fname,6);
		fseek (fout,1,SEEK_SET);
		fwrite ("E",1,1,fout);
		//fseek  (fout,-2,SEEK_CUR);	
		rewind (fout);
		fwrite ("H",1,1,fout);
		fseek (fout,1,SEEK_CUR);
		fwrite ("L",1,1,fout);
		int ret = fclose (fout);
		if (ret!=0){
			t_error ("fclose w(%s)=%d",fname,ret);
		}else{
			FILE *fin = fopen (fname,"r");
			if (fin == NULL){
				t_error ("fopen(%s r)",fname);
			}else{
				char buf[10];
				int len = fread (buf,1,sizeof(buf),fin);
				if (len >= 0) buf[len] = '\0';
				if (len != 6){
					t_error ("fread(%s) len=%d",fname,len);
				}else if (strcmp(buf,"HELlo\n")!=0){
					t_error ("fread(%s) strcmp :%s:",fname,buf);
				}else{
					ret = fclose (fin);
					if (ret!=0){
						t_error ("fclose r(%s)=%d",fname,ret);
					}else{
						unlink (fname);
						t_exist (fname,false);
						t_success();
					}
				}
			}
		}
	}
}

static void t_open (const char *fname)
{
	printf ("Test open: %s\n",fname);
	int fout = open (fname,O_WRONLY|O_CREAT,0644);
	if (fout <= 0){
		t_error ("open(%s O_WRONLY)",fname);
	}else{
		t_exist (fname,true);
		write (fout,"hello\n",6);
		t_size (fname,6);
		lseek (fout,1,SEEK_SET);
		write (fout,"E",1);
		lseek  (fout,-2,SEEK_CUR);	
		write (fout,"H",1);
		lseek (fout,1,SEEK_CUR);
		write (fout,"L",1);
		int ret = close (fout);
		if (ret!=0){
			t_error ("close O_WRONLY(%s)=%d",fname,ret);
		}else{
			int fin = open (fname,O_RDONLY);
			if (fin <= 0){
				t_error ("open(%s O_RDONLY)",fname);
			}else{
				char buf[10];
				int len = read (fin,buf,10);
				if (len >= 0) buf[len] = '\0';
				if (len != 6){
					t_error ("read(%s) len=%d",fname,len);
				}else if (strcmp(buf,"HELlo\n")!=0){
					t_error ("read(%s) strcmp :%s:",fname,buf);
				}else{
					ret = close (fin);
					if (ret!=0){
						t_error ("close O_RDONLY(%s)=%d",fname,ret);
					}else{
						unlink (fname);
						t_exist (fname,false);
						t_success();
					}
				}
			}
		}
	}
}

static void t_dir (const char *dir)
{
	t_exist (dir,false);
	mkdir (dir,0700);
	t_exist (dir,true);
	rmdir (dir);
	t_exist (dir,false);
	t_success();
}

static void t_opendir (const char *dir)
{
	DIR *d = opendir (dir);
	if (d == NULL){
		t_error ("opendir");
	}else{
		struct dirent *ent;
		while ((ent = readdir(d))!=NULL){
			char path[PATH_MAX];
			snprintf (path,sizeof(path)-1,"%s/%s",dir,ent->d_name);
			struct stat st;
			if (stat(path,&st)==-1){
				t_error ("opendir stat %s",ent->d_name);
			}else{
				printf ("entry :%s: uid=%d gid=%d mode=%03o\n",ent->d_name
					,st.st_uid,st.st_gid,st.st_mode);
			}
		}
		closedir (d);
		t_success();
	}
}

static void t_stat (const char *fname)
{
	printf ("test stat\n");
	struct stat st;
	if (stat(fname,&st)==-1){
		t_error ("t_stat");
	}else{
		printf ("\tdev     %02x-%02x\n",major(st.st_dev),minor(st.st_dev));
		printf ("\tino     %ld\n",st.st_ino);
		printf ("\tmode    %04o\n",st.st_mode);
		printf ("\tnlink   %d\n",st.st_nlink);
		printf ("\tuid     %d\n",st.st_uid);
		printf ("\tgid     %d\n",st.st_gid);
		printf ("\trdev    %02x-%02x\n",major(st.st_rdev),minor(st.st_rdev));
		printf ("\tsize    %ld\n",st.st_size);
		printf ("\tblksize %ld\n",st.st_blksize);
		printf ("\tblocks  %ld\n",st.st_blocks);
		printf ("\tatime   %ld\n",st.st_atime);
		printf ("\tmtime   %ld\n",st.st_mtime);
		printf ("\tctime   %ld\n",st.st_ctime);
		t_success();
	}
}
static void t_fchmod (const char *fname)
{
	printf ("test fchmod %s\n",fname);
	int fout = open (fname,O_WRONLY|O_CREAT,0644);
	if (fout <= 0){
		t_error ("open(%s O_WRONLY)",fname);
	}else{
		t_exist (fname,true);
		struct stat st;
		if (stat(fname,&st)==-1 || st.st_mode != 0100644){
			t_error ("stat 0644 %s -> %04o\n",fname,st.st_mode);
		}else if (fchmod (fout,0600)==-1){
			t_error ("fchmod %d errno %d\n",fout,errno);
		}else if (stat(fname,&st)==-1 || st.st_mode != 0100600){
			t_error ("stat 0600 %s -> %04o\n",fname,st.st_mode);
		}
		if (close (fout) == -1){
			t_error ("close %s -> %s\n",fname,errno);
		}
		unlink (fname);
		t_exist (fname,false);
		t_success();
	}
}

static void t_utimes (const char *fname)
{
	printf ("test utimes %s\n",fname);
	int fout = open (fname,O_WRONLY|O_CREAT,0644);
	if (fout <= 0){
		t_error ("open(%s O_WRONLY)",fname);
	}else{
		t_exist (fname,true);
		struct timeval tvp[2];
		tvp[0].tv_sec = 0;
		tvp[1].tv_sec = 0;
		tvp[0].tv_usec = 0;
		tvp[1].tv_usec = 0;
		struct stat st,st2;
		if (stat(fname,&st)==-1){
			t_error ("stat %s\n",fname);
		}else if (utimes (fname,tvp)==-1){
			t_error ("utimes\n");
		}else if (stat(fname,&st2)==-1){
			t_error ("stat %s\n",fname);
		}else if (st2.st_atime != 0 || st2.st_mtime != 0){
			t_error ("atime = %ld, mtime = %ld\n",st2.st_atime
				,st2.st_mtime);
		}else{
			tvp[0].tv_sec = st.st_atime;
			tvp[1].tv_sec = st.st_mtime;
			if (utimes (fname,tvp)==-1){
				t_error ("utimes, second time\n");
			}else if (stat(fname,&st2)==-1){
				t_error ("stat2 %s\n",fname);
			}else if (st2.st_atime != st.st_atime
				|| st2.st_mtime != st.st_mtime){
				t_error ("atime = %ld, mtime = %ld\n",st2.st_atime
					,st2.st_mtime);
			}else{
				unlink (fname);
				t_exist (fname,false);
				t_success();
			}
		}
	}
}

static void t_utime (const char *fname)
{
	printf ("test utimes %s\n",fname);
	int fout = open (fname,O_WRONLY|O_CREAT,0644);
	if (fout <= 0){
		t_error ("open(%s O_WRONLY)",fname);
	}else{
		t_exist (fname,true);
		struct utimbuf tvp;
		tvp.actime = 0;
		tvp.modtime = 0;
		struct stat st,st2;
		if (stat(fname,&st)==-1){
			t_error ("stat %s\n",fname);
		}else if (utime (fname,&tvp)==-1){
			t_error ("utime\n");
		}else if (stat(fname,&st2)==-1){
			t_error ("stat %s\n",fname);
		}else if (st2.st_atime != 0 || st2.st_mtime != 0){
			t_error ("atime = %ld, mtime = %ld\n",st2.st_atime
				,st2.st_mtime);
		}else{
			tvp.actime = st.st_atime;
			tvp.modtime = st.st_mtime;
			if (utime (fname,&tvp)==-1){
				t_error ("utime, second time\n");
			}else if (stat(fname,&st2)==-1){
				t_error ("stat2 %s\n",fname);
			}else if (st2.st_atime != st.st_atime
				|| st2.st_mtime != st.st_mtime){
				t_error ("atime = %ld, mtime = %ld\n",st2.st_atime
					,st2.st_mtime);
			}else{
				unlink (fname);
				t_exist (fname,false);
				t_success();
			}
		}
	}
}

#if __GLIBC__ > 2 || ( __GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)

static void t_stat64 (const char *fname)
{
	#ifdef __USE_LARGEFILE64
		printf ("test stat64\n");
		struct stat64 st;
		if (stat64(fname,&st)==-1){
			t_error ("t_stat");
		}else{
			printf ("\tdev     %02x-%02x\n",major(st.st_dev),minor(st.st_dev));
			printf ("\tino     %ld\n",st.st_ino);
			printf ("\tmode    %04o\n",st.st_mode);
			printf ("\tnlink   %d\n",st.st_nlink);
			printf ("\tuid     %d\n",st.st_uid);
			printf ("\tgid     %d\n",st.st_gid);
			printf ("\trdev    %02x-%02x\n",major(st.st_rdev),minor(st.st_rdev));
			printf ("\tsize    %Ld\n",st.st_size);
			printf ("\tblksize %ld\n",st.st_blksize);
			printf ("\tblocks  %Ld\n",st.st_blocks);
			printf ("\tatime   %ld\n",st.st_atime);
			printf ("\tmtime   %ld\n",st.st_mtime);
			printf ("\tctime   %ld\n",st.st_ctime);
			t_success();
		}
	#else
		fprintf (stderr,"Not compiled with stat64 support\n");
	#endif
}

#endif

static void t_assert_absolute(const char *dir)
{
	const char *end = strrchr(dir,'/');
	if (end == NULL){
		t_error ("path must be absolute");
	}
}

static void t_create (const char *fname)
{
	FILE *fout = fopen (fname,"w");
	if (fout != NULL) fclose (fout);
	t_exist (fname,true);
}

static char *t_getcwd(char *cwd, int size)
{
	printf ("Before getcwd\n");
	strcpy (cwd,"Hello");
	char *ret = getcwd(cwd,size);
	printf ("After getcwd :%s:\n",cwd);
	return ret;
}
	

static void t_chdir (const char *dir)
{
	printf ("test chdir\n");
	const char *end = strrchr(dir,'/');
	char cwd[PATH_MAX];
	t_assert_absolute(dir);
	if (chdir (dir)!=0){
		t_error ("chdir %s",dir);
	}else if (t_getcwd(cwd,sizeof(cwd)-1)==NULL){
		t_error ("getcwd");
	}else if (strcmp(cwd,dir)!=0){
		t_error ("strcmp :%s: :%s: != 0",dir,cwd);
	}else if (chdir ("..")!=0){
		t_error ("chdir ..");
	}else if (chdir(end+1)!=0){
		t_error ("chdir %s",end+1);
	}else{
		t_success();
	}
}

static void t_rename (const char *dir)
{
	printf ("test rename [%s]\n",dir);
	t_assert_absolute(dir);
	char src[PATH_MAX],dst[PATH_MAX],sub[PATH_MAX];
	sprintf (sub,"%s/renamedir",dir);
	sprintf (src,"%s/renamedir/toto.old",dir);
	sprintf (dst,"%s/renamedir/toto.new",dir);
	t_exist (sub,false);
	mkdir   (sub,0755);
	t_exist (sub,true);
	t_create (src);
	unlink (dst);
	if (rename(src,dst)==-1){
		t_error ("rename %s -> %s",src,dst);
	}else{
		t_exist (dst,true);
		unlink (dst);
		t_exist (dst,false);
		rmdir   (sub);
		t_exist (sub,false);
		t_success ();
	}
}

static void t_link (const char *dir)
{
	printf ("test link\n");
	t_assert_absolute(dir);
	char src[PATH_MAX],dst[PATH_MAX];
	sprintf (src,"%s/toto.old",dir);
	sprintf (dst,"%s/toto.new",dir);
	t_create (src);
	unlink (dst);
	if (link(src,dst)==-1){
		t_error ("link %s -> %s",src,dst);
	}else{
		t_exist (src,true);
		t_exist (dst,true);
		unlink (src);
		t_exist (src,false);
		t_exist (dst,true);
		unlink (dst);
		t_exist (dst,false);
		t_success ();
	}
}

static void t_symlink (const char *dir)
{
	printf ("test symlink\n");
	t_assert_absolute(dir);
	char src[PATH_MAX],dst[PATH_MAX];
	sprintf (src,"%s/toto.old",dir);
	sprintf (dst,"%s/toto.new",dir);
	t_create (src);
	unlink (dst);
	if (symlink("toto.old",dst)==-1){
		t_error ("symlink %s -> %s",src,dst);
	}else{
		t_exist (src,true);
		t_exist (dst,true);
		unlink (src);
		t_exist (src,false);
		t_exist (dst,false);
		unlink (dst);
		t_exist (dst,false);
		t_success ();
	}
}


int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		fprintf (stderr,"fstest option value [ option value ]\n");
	}else{
		ret = 0;
		for (int i=1; i<argc; i++){
			const char *arg1 = argv[i];
			if (strcmp(arg1,"--fopen")==0){
				t_fopen (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--open")==0){
				t_open (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--dir")==0){
				t_dir (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--opendir")==0){
				t_opendir (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--chdir")==0){
				t_chdir (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--rename")==0){
				t_rename (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--link")==0){
				t_link (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--symlink")==0){
				t_symlink (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--stat")==0){
				t_stat (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--fchmod")==0){
				t_fchmod (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--utimes")==0){
				t_utimes (argv[i+1]);
				i++;
			}else if (strcmp(arg1,"--utime")==0){
				t_utime (argv[i+1]);
				i++;
			#if __GLIBC__ > 2 || ( __GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)
				}else if (strcmp(arg1,"--stat64")==0){
					t_stat64 (argv[i+1]);
					i++;
			#endif
			}else{
				fprintf (stderr,"Invalid option %s\n",arg1);
				ret = -1;
				break;
			}
		}
	}
	return ret;
}

