/*
	Tell a back end to terminate
*/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"

static void usage()
{
	fprintf (stderr,"litefs-shutdown version 1.0\n"
		"\n"
		"litefs-shutdown --sockets unix_socket [ unix_socket ]\n");
}

static int litefs_shutdown (const char *sockpath, bool showerror)
{
	int ret = -1;
	int fd =  socket (AF_UNIX,SOCK_STREAM,0);
	if (fd == -1){
		if (showerror) perror("socket client");
	}else{
		struct sockaddr_un un;
		un.sun_family = AF_UNIX;
		strcpy (un.sun_path,sockpath);
		int s = connect(fd,(struct sockaddr*)&un,sizeof(un));
		if (s == -1){
			if (showerror) perror("connect");
		}else{
			LITE_REQUEST req;
			req.cmd = LITE_SHUTDOWN;
			req.mode = 0;
			req.perm = 0;
			req.len  = 0;
			int ret = write (fd,&req,sizeof(LITE_REQUEST));
			LITE_REPLY rep;
			int len = read (fd,&rep,sizeof(rep));
			if (len == sizeof(LITE_REPLY)
				&& rep.status == LITE_STATUS_OK){
				ret = 0;
			}
		}
		close (fd);
	}
	return ret;
}


int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--sockets")==0){
				i++;
				for (; i<argc; i++){
					litefs_shutdown (argv[i],false);
				}
				ret = 0;
			}else{
				usage();
			}
		}
	}
	return ret;
}	

