#ifndef LITEFS_H
#define LITEFS_H

#include <sys/stat.h>
#include <limits.h>

#define LITEFS_PROTOCOL_VERSION	5

enum LITE_CMD{
	LITE_OPEN,
	LITE_CLOSE,
	LITE_READ,
	LITE_WRITE,
	LITE_SEEK,
	LITE_STAT,
	LITE_LSTAT,
	LITE_FSTAT,
	LITE_OPENDIR,
	LITE_READDIR,
	LITE_CLOSEDIR,
	LITE_UNLINK,
	LITE_RENAME,
	LITE_MKDIR,
	LITE_RMDIR,
	LITE_MKNOD,
	LITE_LINK,
	LITE_SYMLINK,
	LITE_CHMOD,
	LITE_CHOWN,
	LITE_LCHOWN,
	LITE_CHDIR,
	LITE_UTIMES,
	LITE_ACCESS,
	LITE_READLINK,
	LITE_VERSION,
	LITE_SHUTDOWN,		// End litefs back-end
	LITE_FCHMOD,
	LITE_FCHOWN,
	// Stuff for remote execution
	LITE_EXEC,			// Remote execution on the back-end
	LITE_EXECIN,		// Send standard input to a running program
	LITE_SIGNAL,		// A signal (kill) is sent to the remote process
	LITE_SETWIN,		// Send new window settings
						// data contain LITE_WININFO
};

struct LITE_REQUEST {
	short cmd;
	short handle;
	long mode;
	long perm;
	short client_handle;	// Used by litefs-proxy to demultiplex
							// the requests
	unsigned short len;		// Len of data
};

struct LITE_REQUEST_DATA: LITE_REQUEST {
	char buf[32000];
};

struct LITE_WININFO {
	short width;
	short height;
};

struct LITE_EXECINFO {
	LITE_WININFO win;
	short isatty0;		// Is handle 0 a tty
	short nbenv;
	short nbarg;
};

struct LITE_EXECINFO_DATA: LITE_EXECINFO {
	char data[30000];
};
enum LITE_STATUS {
	LITE_STATUS_FAIL,		// The request failed
	LITE_STATUS_OK,			// The request succeed, no special value
							// returned
	LITE_STATUS_HANDLE,		// A file handle was transmited
	LITE_STATUS_STAT,		// struc stat was sent
	LITE_STATUS_SEEK,		// data contains lseek return value
	LITE_PDATA_OUT,			// output from a running program
	LITE_PDATA_ERR,			// error from a running program
	LITE_STATUS_EXEC,		// Return code of the remote process
	LITE_STATUS_TERM,		// Termio info (remote execution stuff)
							// Data contain LITE_TERMIO
};

struct LITE_TERMIO{
};

struct LITE_REPLY {
	short status;
	short errcode;
	int data;
	short client_handle;
	unsigned short len;
};

struct LITE_REPLY_DATA: LITE_REPLY {
	char buf[32000];
};

struct LITEFSD_CTX{
	int fd;
	int client_handle;
};

#endif


