/*
	This is a personnal automounter.
	It does some automounting/remounting based on user id.

	It also does a fancy tricks. It simulates the mounts inside
	$HOME/n/..., but effectivly does it in $HOME/.n/... The goal
	is to allow unmounting a volume even if the current working directory
	is inside that volume.

	It is also called a personnal automounter because it should manage
	personnal network mounts such as SMB mounts. When doing a personnal
	mount, it will prompt for a password and use a partner program to
	remember the password for the duration of the local session.
	At some point, this partner will be able to encrypt the password
	and use the local password as the key.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <netinet/in.h>
#include "remaudio.h"
#include "dispatch.h"


static int persmount_mkdiranc (const char *path)
{
	int ret = __mkdir (path,0755);
	if (ret == -1){
		char tmp[strlen(path)+1];
		strcpy (tmp,path);
		char *pt = strrchr(tmp,'/');
		if (pt != NULL){
			*pt = '\0';
			if (persmount_mkdiranc (tmp) != -1) ret = __mkdir (path,0755);
		}
	}
	return ret;
}


/*
	Make sure a mount point is mounted.
	If not, execute the proper command
*/
static void domount(const char *mpoint, const char *command)
{
	FILE *fin = fopen ("/proc/mounts","r");
	if (fin != NULL){
		bool is_mounted = false;
		char buf[2*PATH_MAX];
		while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
			char dev[PATH_MAX],dir[PATH_MAX],stuff[PATH_MAX];
			if (sscanf (buf,"%s %s %s",dev,dir,stuff)>=2
				&& strcmp(dir,mpoint)==0){
				is_mounted = true;
				break;
			}
		}
		fclose (fin);
		if (!is_mounted){
			persmount_mkdiranc (mpoint);
			// fprintf (stderr,"mounting :%s:\n",mpoint);
			_dispatch_exec(command);
		}
	}
}

struct PERSMOUNT_INFO{
	char volume[PATH_MAX];		// Sub-directory in $HOME/n
	char relpath[PATH_MAX];		// relative path in $HOME/n/volume/
	char virtpath[PATH_MAX];	// absolute virtual path
	char realpath[PATH_MAX];	// Translated path
};

struct PERSMOUNT_CONF {
	char *volume;		// Name of the virtual mount point in $HOME/n
	char *mountpoint;	// Absolute path of the mount point
						// Not always a mount point. If there is no
						// command, then it is not a mount poing
	char *command;		// Optional command to do the mount
	time_t lastcheck;	// We do not check that the mount is effective
						// at every request. We assume that once a mount
						// is done, it stands for a least 3 seconds
};

static int nbdef=0;
static PERSMOUNT_CONF *conf;

static void persmount_readconf ()
{
	if (conf == NULL){
		char path[PATH_MAX];
		snprintf (path,sizeof(path)-1,"%s/.persmount",_dispatch_gethome());
		FILE *fin = fopen (path,"r");
		if (fin != NULL){
			char buf[1000];
			conf = (PERSMOUNT_CONF*)malloc (100*sizeof(PERSMOUNT_CONF));
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				char *start = buf;
				while (isspace(*start)) start++;
				if (start[0] != '\0' && start[0] != '#'){
					char *pt = start;
					while (*pt > ' ') pt++;
					if (isspace(*pt)){
						*pt ++ = '\0';
						while (isspace(*pt)) pt++;
						char *mpt = pt;
						while (*pt > ' ') pt++;
						if (isspace(*pt)) *pt++ = '\0';
						while (isspace(*pt)) pt++;
						char *end = pt + strlen(pt)-1;
						while (*end <= ' ' && end >= pt) *end-- = '\0';
						conf[nbdef].volume = strdup(start);
						if (mpt[0] == '/'){
							conf[nbdef].mountpoint = strdup(mpt);
						}else{
							// Mount point is relative to home
							char absmpt[PATH_MAX];
							snprintf (absmpt,sizeof(absmpt)-1,"%s/%s"
								,_dispatch_gethome(),mpt);
							conf[nbdef].mountpoint = strdup(absmpt);
						}
						if (pt[0] == '\0'){
							conf[nbdef].command = NULL;
						}else{
							conf[nbdef].command = strdup(pt);
						}
						conf[nbdef].lastcheck = 0;
						nbdef++;
					}
				}
			}
			fclose (fin);
		}
	}
}

/*
	Check if a volume is configured for the personal auto mounter
	and mount it if needed.
*/
static int persmount_chkconf (PERSMOUNT_INFO &info, bool only_check)
{
	persmount_readconf();
	int ret = -1;
	PERSMOUNT_CONF *ptc = conf;
	for (int i=0; i<nbdef; i++,ptc++){
		if (strcmp(info.volume,ptc->volume)==0){
			strcpy (info.realpath,ptc->mountpoint);
			if (!only_check && ptc->command != NULL){
				time_t now = time(NULL);
				if (now > ptc->lastcheck){
					domount(info.realpath,ptc->command);
					ptc->lastcheck = now + 2;
				}
			}
			strcat (info.realpath,info.relpath);
			ret = 0;
			break;
		}
	}
	return ret;
}
static void domount_remote (const char *volume)
{
	static time_t lasttime=0;
	time_t now = time(NULL);
	// Avoid sending too many request to the remaudioserver.
	if (now - lasttime > 2){
		int fd_remote= __clientsock_connect (__audioserver,8010,1);
		if (fd_remote != -1){	
			// Send a mount request to make sure the volume is still
			// mounted on the remote side. On our side, we do an NFS
			// mount and keeps it. We will probably clean that at
			// login time or logout time.
			REMAUDIO_OPEN op;
			op.version = htons (REMAUDIO_VERSION);
			op.cmd = htons(REMAUDIO_MOUNT);
			strcpy (op.fname,volume);
			int uid = getuid();
			op.mode = htonl(uid);
			__libc_write (fd_remote,&op,sizeof(op));
			REMAUDIO_CTL ctl;
			__libc_read(fd_remote,&ctl,sizeof(ctl));	// We do not care about the
												// the return code.
			__libc_close (fd_remote);
			lasttime = time(NULL);
		}
	}
}
#if 0
static void domount_remote_nfs (const char *volume, char *realpath)
{
	snprintf (realpath,PATH_MAX-1,"/mnt/%s/%s",__audioserver,volume);
	char cmd[PATH_MAX];
	snprintf (cmd,sizeof(cmd)-1,"/usr/sbin/persmount_remote %s %s"
		,__audioserver,volume);
	domount (realpath,cmd);
	domount_remote (volume);
}
#endif


/*
	Translate a logical path if it is part of the personnal mount area
	($HOME/n)
*/
static int persmount_translate (
	const char *path,
	PERSMOUNT_INFO &info,
	bool mount_anyway)	// Do the mount even if we are not inside the mount
						// point
{
	int lenn;
	const char *pathn = _dispatch_getpathn(lenn);
	if (pathn == NULL) return -1;
	strcpy (info.realpath,path);
	if (path[0] != '/'){
		const char *curpath = _dispatch_getcwd();
		snprintf (info.virtpath,sizeof(info.virtpath)-1,"%s/%s",curpath,path);
		path = info.virtpath;
	}else{
		strcpy (info.virtpath,path);
	}
	int ret = -1;

	// fprintf (stderr,"pathn = :%s: <> :%s: cwd(%s)\n",pathn,path,_dispatch_getcwd());
	if (strncmp(pathn,path,lenn)==0){
		path += lenn;
		char *volume = info.volume;
		while (*path != '\0' && *path != '/'){
			*volume++ = *path++;
		}
		*volume = '\0';
		strcpy (info.relpath,path);
		#if 0
			fprintf (stderr,"translate :%s: :%s:\n",info.volume,info.relpath);
			fflush (stderr);
		#endif
		/* #Specification: persmount / trigger
			We only trigger the mount when accessing files inside the
			mount point or when a specific directory operation is
			used on the mount point (opendir). Most software (ls, file
			browsers, ...) do a trick to deal with symlinks. They do a
			stat syscall on path/. instead of just path. To avoid
			triggering a mount of every personal volumes, we manage
			path/. as path (so not inside the mount point).
		*/
		if (!(path[0] == '\0'
			|| (path[0] == '/'
				&& path[1] == '.'
				&& path[2] == '\0'))){
			mount_anyway = true;
		}
		// fprintf (stderr,"persmount volume :%s: sub :%s: anyway %d\n",info.volume,path,mount_anyway);
		ret = 0;
		if (strcmp(info.volume,"a:")==0){
			if (__remaudio_isremote()){
				// This code is not executed anymore, because litefs intercept
				// the request
				//if (mount_anyway) domount_remote_nfs ("floppy",info.realpath);
				ret = -1;
			}else{
				if (mount_anyway) domount_remote ("floppy");
				strcpy (info.realpath,"/mnt/floppy");
			}
			// fprintf (stderr,"floppy :%s: :%s:\n",info.realpath,info.relpath);
			strcat (info.realpath,info.relpath);
		}else if (strcmp(info.volume,"b:")==0){
			if (__remaudio_isremote()){
				//if (mount_anyway) domount_remote_nfs ("fd1",info.realpath);
				ret = -1;
			}else{
				if (mount_anyway) domount_remote ("fd1");
				strcpy (info.realpath,"/mnt/fd1");
			}
			strcat (info.realpath,info.relpath);
		}else if (strcmp(info.volume,"cdrom")==0){
			if (__remaudio_isremote()){
				//if (mount_anyway) domount_remote_nfs ("cdrom",info.realpath);
				ret = -1;
			}else{
				if (mount_anyway) domount_remote ("cdrom");
				strcpy (info.realpath,"/mnt/cdrom");
			}
			strcat (info.realpath,info.relpath);
		}else{
			if (persmount_chkconf (info,!mount_anyway)==-1){
				// When it fails, we reread the .persmount file
				// in case it was reconfigured

				// Note that this does not kill too much performance
				// because persmount is loaded first (after remaudio)
				// so end up last in the linked list of virtualfs driver
				// so the request seldom fail.
				for (int i=0; i<nbdef; i++){
					free (conf[i].volume);
					free (conf[i].command);
				}
				free (conf);
				nbdef = 0;
				conf = NULL;
				ret = persmount_chkconf (info,!mount_anyway);
			}
		}
	}
	return ret;
}

#if 0
static int checkmount (const char *path)
{
	// Check mount is itself opening file. The open is trapped in
	// remaudio.cc and itself call checkmount. So we protect against
	// reentrancy. This is simpler then forcing function below to
	// to use lower level function.
	int ret = -1;
	static bool checkmounting = false;
	if (!checkmounting){
		checkmounting = true;
		PERSMOUNT_INFO info;
		persmount_translate (path,info,false);
		checkmounting = false;
	}
	return ret;
}

static AUTOMOUNT p={
	NULL,
	checkmount,
	NULL,			// afterchdir
};

static REGISTER_AUTOMOUNT register_autom (p);

#endif

static int persmount_open (const char *path, int mode, int perm)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (path,info,false)!=-1){
		// fprintf (stderr,"open :%s: %d %d\n",info.realpath,mode,perm);
		ret = open (info.realpath,mode,perm);
	}
	return ret;
}


static int persmount_chdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (path,info,false)!=-1){
		// fprintf (stderr,"chdir :%s: :%s: :%s:\n",path,info.virtpath,info.realpath);
		ret = __chdir (info.realpath);
		if (ret != -1){
			// We have to set the new current dir in $HOME/n
			if (__chdir (info.virtpath) == -1){
				persmount_mkdiranc (info.virtpath);
				__chdir (info.virtpath);
			}
		}
	}
	return ret;
}

static struct DIR * persmount_opendir (const char *path)
{
	struct DIR *ret = NULL;
	PERSMOUNT_INFO info;
	if (persmount_translate (path,info,true)!=-1){
		// fprintf (stderr,"opendir :%s:\n",info.realpath);
		ret = __opendir (info.realpath);
	}
	return ret;
}

static struct dirent *persmount_readdir (struct DIR *dir)
{
	return __readdir (dir);
}


static int persmount_closedir (struct DIR *dir)
{
	return __closedir (dir);
}
static int persmount_rename (const char *old, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO oldinfo;
	PERSMOUNT_INFO newinfo;
	int oldf = persmount_translate (old,oldinfo,false);
	int newf = persmount_translate (newp,newinfo,false);
	if (oldf !=-1 && newf != -1){
		ret = __rename (oldinfo.realpath,newinfo.realpath);
	}else if (oldf != -1){
		ret = __rename (oldinfo.realpath,newp);
	}else if (newf != -1){
		ret = __rename (old,newinfo.realpath);
	}
	return ret;
}
static int persmount_unlink (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (path,info,false)!=-1){
		ret = __unlink (info.realpath);
	}
	return ret;
}

static int persmount_link (const char *oldpath, const char *newpath)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO oldinfo,newinfo;
	if (persmount_translate (oldpath,oldinfo,false)!=-1
		|| persmount_translate (newpath,newinfo,false)!=-1){
		ret = __link (oldinfo.realpath,newinfo.realpath);
	}
	return ret;
}

static int persmount_symlink (const char *oldpath, const char *newpath)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO newinfo;
	if (persmount_translate (newpath,newinfo,false)!=-1){
		ret = __symlink (oldpath,newinfo.realpath);
	}
	return ret;
}
static int persmount_rmdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (path,info,false)!=-1){
		ret = __rmdir (info.realpath);
	}
	return ret;
}
static int persmount_mkdir(const char *path, int mode)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (path,info,false)!=-1){
		ret = __mkdir (info.realpath,mode);
	}
	return ret;
}

static int persmount_stat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		// fprintf (stderr,"persmount_stat :%s:\n",info.realpath);
		ret = __xstat (ver,info.realpath,buf);
	}
	return ret;
}
static int persmount_lstat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		// fprintf (stderr,"persmount_lstat :%s:\n",info.realpath);
		ret = __lxstat (ver,info.realpath,buf);
	}
	return ret;
}

static int persmount_chmod (const char *fname, mode_t mode)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		// fprintf (stderr,"chmod :%s: %d\n",info.realpath,mode);
		ret = __chmod (info.realpath,mode);
	}
	return ret;
}
static int persmount_chown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		// fprintf (stderr,"chown :%s: %d %d\n",info.realpath,uid,gid);
		ret = __chown (info.realpath,uid,gid);
	}
	return ret;
}
static int persmount_lchown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		// fprintf (stderr,"lchown :%s: %d %d\n",info.realpath,uid,gid);
		ret = __lchown (info.realpath,uid,gid);
	}
	return ret;
}

static int persmount_readlink (const char *fname, char *buf, size_t bufsiz)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		ret = __readlink (info.realpath,buf,bufsiz);
	}
	return ret;
}
#if 0
static int persmount_utimes (const char *fname, struct timeval *tvp)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	// fprintf (stderr,"utimes :%s:\n",info.realpath);
	if (persmount_translate (fname,info,false)!=-1){
		// fprintf (stderr,"utimes :%s:\n",info.realpath);
		ret = __utimes (info.realpath,tvp);
	}
	return ret;
}
#endif
static int persmount_utime (const char *fname, const struct utimbuf *tvp)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	// fprintf (stderr,"utimes :%s:\n",info.realpath);
	if (persmount_translate (fname,info,false)!=-1){
		// fprintf (stderr,"utimes :%s:\n",info.realpath);
		ret = __utime (info.realpath,tvp);
	}
	return ret;
}

static int persmount_access (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		ret = __access (info.realpath,type);
	}
	return ret;
}
static int persmount_euidaccess (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (fname,info,false)!=-1){
		ret = __euidaccess (info.realpath,type);
	}
	return ret;
}

static int persmount_execve (
	const char *filename,
	char *const argv[],
	char *const envp[])
{
	int ret = DISPATCH_DONTCARE;
	PERSMOUNT_INFO info;
	if (persmount_translate (filename,info,false)!=-1){
		ret = __execve (info.realpath,argv,envp);
	}
	return ret;
}



static DISPATCH m={
	NULL,
	persmount_open,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	persmount_stat,
	NULL,
	persmount_lstat,
	persmount_chdir,
	persmount_opendir,
	persmount_readdir,
	persmount_closedir,
	persmount_rename,
	persmount_unlink,
	persmount_rmdir,
	persmount_mkdir,
	NULL,		// socket
	NULL,		// connect
	NULL,		// bind
	persmount_chmod,
	persmount_chown,
	persmount_lchown,
	persmount_readlink,
	persmount_utime,
	persmount_access,
	persmount_euidaccess,
	persmount_link,
	persmount_symlink,
	NULL,		// fchmod
	NULL,		// fchown
	persmount_execve,
};

static REGISTER_DISPATCH register_m (m);

